#pragma once

/* libc includes */
#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>

/* Genode includes */
#include <base/heap.h>
#include <base/mutex.h>
#include <libc/component.h>
#include <timer_session/connection.h>
#include <util/list.h>

/* gapfruit includes */
#include <picosha2/picosha2.h>
#include <terminal_session/connection.h>
#include <util/hexdump.h>

/* local includes */
#include "command.h"
#include "crypto_enums.h"
#include "error.h"
#include "exception.h"
#include "message.h"
#include "payload.h"
#include "socket_wrapper.h"


namespace Hsm {
	namespace internal {
    	template<typename SOCKET_TYPE, typename RANDOM_TERMINAL>
		class Connection;
	}

	using Connection = internal::Connection<Socket_wrapper, Terminal::Connection>;

	using Hostname   = Genode::String<256>;
	using Ip_address = Genode::String<64>;
	using Username   = Genode::String<256>;
	using Password   = Genode::String<256>;
	using Key_name   = Genode::String<256>;
}

template<typename SOCKET_TYPE, typename RANDOM_TERMINAL>
class Hsm::internal::Connection
{
	private:
		using Socket         = Genode::Constructible<SOCKET_TYPE>;
		using Signal_handler = Genode::Signal_handler<Connection>;

		enum {
			INVALID_SOCKET = -1,
			SALT_SIZE = 32,
			HASH_SIZE = 32,
			MAGIC_SIZE = 32,
		};

		constexpr static uint8_t const magic_login_request[32] = {
			0x61, 0x8d, 0x63, 0xaa,
			0xe8, 0x89, 0xca, 0x1e,
			0x6c, 0x79, 0x1e, 0xb2,
			0x7d, 0xfd, 0x4a, 0xd5,
			0x03, 0x74, 0x6e, 0x7f,
			0xa0, 0xc1, 0x8e, 0xd3,
			0x51, 0xb5, 0x5d, 0x71,
			0xdb, 0x3f, 0x4f, 0x48 };
		constexpr static uint8_t const magic_login_auth[32] = {
			0xda, 0x7e, 0xa8, 0x27,
			0xe4, 0xae, 0xfb, 0x67,
			0x48, 0x36, 0xd0, 0xc2,
			0x1e, 0x32, 0xca, 0x24,
			0xb1, 0xc7, 0x6d, 0xe4,
			0xcf, 0xdb, 0x37, 0x11,
			0x94, 0x07, 0x4e, 0x20,
			0x97, 0x07, 0x8b, 0x98 };

		Genode::Env&           _env;
		Genode::Heap&          _heap;

		RANDOM_TERMINAL&       _random_terminal;
		Username               _username;
		Password               _password;
		mutable bool           _login_success    { false };
		Socket                 _socket           { };
		mutable Genode::Mutex  _config_mutex     { };

		void _login_request(const Username& username,
		                    uint8_t* client_login_salt,
		                    uint8_t* server_login_salt) const
		{
			Message request(_heap, Command::CommandLoginRequest);
			request.add(Data_type::typeIdLabel, username.string());
			_random_terminal.read(client_login_salt, SALT_SIZE);
			request.add(Data_type::typeRandomData, client_login_salt, SALT_SIZE);
			request.add(Data_type::typeUserId, username.string());

			Message const response = execute(request);

			switch (response.status())
			{
				case CKR_OK:
					break;
				default:
					Genode::error("HSM replied with error ", Genode::Hex(response.status()));
					throw Hsm::Error_response_exception();
			}

			Payload const server_salt_payload = response.payload_by_type(Data_type::typeLoginData);
			if (!server_salt_payload.is_valid()) {
				Genode::error("Server salt missing");
				throw Hsm::Missing_payload_exception();
			}
			Genode::memcpy(server_login_salt, server_salt_payload.data(), SALT_SIZE);
		}

		void _compute_challenge(const uint8_t* magic,
		                        const uint8_t* salt1,
		                        const uint8_t* salt2,
		                        const Password& password,
		                        uint8_t* challenge) const
		{
			const char* const pw = password.string();
			std::vector<uint8_t> pw_hash(HASH_SIZE);
			picosha2::hash256(pw, pw + strlen(pw), pw_hash.begin(), pw_hash.end());

			picosha2::hash256_one_by_one hasher;
			hasher.process(magic, magic + MAGIC_SIZE);
			hasher.process(salt1, salt1 + SALT_SIZE);
			hasher.process(salt2, salt2 + SALT_SIZE);
			hasher.process(pw_hash.begin(), pw_hash.end());
			hasher.finish();

			hasher.get_hash_bytes(challenge, challenge + HASH_SIZE);
		}

		bool _login_response(const Password& password,
		                     const uint8_t* client_login_salt,
		                     const uint8_t* server_login_salt) const
		{
			uint8_t request_hash[HASH_SIZE];
			_compute_challenge(magic_login_request, server_login_salt,
			                   client_login_salt, password, request_hash);

			Message request(_heap, Command::CommandLoginResponse);
			request.add(Data_type::typeLoginData, request_hash, sizeof(request_hash));
			uint8_t client_auth_salt[SALT_SIZE];
			_random_terminal.read(client_auth_salt, sizeof(client_auth_salt));
			request.add(Data_type::typeRandomData, client_auth_salt, sizeof(client_auth_salt));

			Message const response = execute(request);

			switch (response.status())
			{
				case CKR_OK:
					break;
				default:
					Genode::error("HSM replied with error ", Genode::Hex(response.status()));
					throw Hsm::Error_response_exception();
			}

			Payload const server_salt_payload = response.payload_by_type(Data_type::typeRandomData);
			if (!server_salt_payload.is_valid()) {
				Genode::error("Server salt missing");
				throw Hsm::Missing_payload_exception();
			}
			Payload const client_challenge = response.payload_by_type(Data_type::typeLoginData);
			if (!client_challenge.is_valid()) {
				Genode::error("Client_challenge missing");
				throw Hsm::Missing_payload_exception();
			}
			uint8_t auth_hash[HASH_SIZE];
			_compute_challenge(magic_login_auth, server_salt_payload.data(), client_auth_salt, password, auth_hash);
			return Genode::memcmp(auth_hash, client_challenge.data(), SALT_SIZE) == 0;
		}

		enum class Counter_action {GetValue = 0, IncrementValue = 1};
		bool _user_counter_value(Genode::uint64_t &counter_value, Counter_action action = Counter_action::GetValue) const
		{
			constexpr static Command s_commands[] = {Command::CommandGetUserCounterValue, Command::CommandIncrementUserCounter};
			Message const request(_heap, s_commands[static_cast<int>(action)]);
			Message const response = execute(request);

			if (response.status() != CKR_OK) {
				// this is possible when the HSM doesn't have the counter option enabled
				Genode::error("HSM replied with error ", Genode::Hex(response.status()));
				throw Hsm::Error_response_exception();
			}

			Payload const payload = response.payload_by_type(Data_type::type64BitCounter);
			if (!payload.is_valid()) {
				Genode::warning("counter value data missing");
				return false;
			} else {
				counter_value = payload.value<Genode::uint64_t>();
				return true;
			}
		}

		void _start_login()
		{
			try {
				hello();
				_try_login();
			} catch (Hsm::Error_response_exception&) {
				error("HSM login failed.");
				_socket->close_and_reconnect();
			} catch (Hsm::Max_connections_exceeded&) {
				error("HSM reported: max connection exceeded (status ",
				      ERR_MAX_CONNECTIONS_EXCEEDED, ") reconnecting");
				_socket->close_and_reconnect(false);
			}
		}

		void _try_login() const
		{
			if (_login_success) {
				Genode::log("HSM connected");
				return;
			}

			if (login(_username, _password)) {
				_login_success = true;
				log("HSM login successful.");
			} else {
				error("HSM login failed.");
			}
		}

		void _disconnected()
		{
			Genode::log("HSM disconnected");
			_login_success = false;
		}

	public:

		Connection(Genode::Env& env,
		           Genode::Heap& heap,
		           RANDOM_TERMINAL& random_terminal,
		           const Ip_address& ip_address,
		           const uint16_t port,
		           const Username& username,
		           const Password& password) :
			_env             { env },
			_heap            { heap },
			_random_terminal { random_terminal },
			_username        { username },
			_password        { password }
		{
			_socket.construct(env, ip_address, port,
			                  [&] () { _start_login(); },
			                  [&] () { _disconnected(); });
		}

		bool connected() const { return _socket->connected(); }

		bool is_logged() const { return _login_success; }

		void update_config(const Ip_address& ip_address,
		                   const uint16_t port,
		                   const Username& username,
		                   const Password& password)
		{
			if (password   == _password &&
			    username   == _username &&
			    port       == _socket-> port() &&
			    ip_address == _socket->ip_address()) {
				return;
			}

			Genode::Mutex::Guard guard { _config_mutex };

			Genode::log("HSM configuration changed - disconnecting from old destination");

			_username = username;
			_password = password;

			_socket->change_destination(ip_address, port);
		}

		Message execute(const Message& request) const
		{
			Genode::Mutex::Guard guard { _config_mutex };

			using Genode::Hex;
			_socket->write(request.data(), request.size());
			Message response(_heap, *_socket);

			if (response.status() == ERR_MAX_CONNECTIONS_EXCEEDED) {
				throw Hsm::Max_connections_exceeded();
			} else if (response.status() >= SECUROSYS_ERROR_BASE) {
				Genode::error("HSM replied with error ", Genode::Hex(response.status()));
				throw Hsm::Error_response_exception();
			} else if (response.status() != CKR_OK) {
				Genode::warning("HSM replied with status ", Genode::Hex(response.status()));
			}

			return response;
		}

		void hello() const
		{
			Message request(_heap, Command::CommandHello);
			request.add(Data_type::typeMagic, LOGIN_MAGIC);

			Message const response = execute(request);

			switch (response.status())
			{
				case CKR_OK:
					break;
				default:
					Genode::error("HSM replied with error ", Genode::Hex(response.status()));
					throw Hsm::Error_response_exception();
			}
		}

		bool login(const Username& username, const Password& password) const
		{
			uint8_t client_salt[32];
			uint8_t server_salt[32];
			_login_request(username, client_salt, server_salt);
			return _login_response(password, client_salt, server_salt);
		}

		uint32_t convert_signature(const Key_name& key_name,
		                           const uint8_t* raw_signature_data,
		                           const uint32_t raw_signature_size,
		                           uint8_t* der_signature_data,
		                           const uint32_t der_signature_size) const
		{
			if (!_login_success) {
				throw Not_logged_in { };
			}

			Message request(_heap, Command::CommandEncodeSignature);
			request.add(Data_type::typeCryptoMethod, Crypto_method::metEcDsaSha2256);
			request.add(Data_type::typeSignatureData, raw_signature_data, raw_signature_size);
			request.add(Data_type::typeIdLabel, key_name.string());

			Message const response = execute(request);

			switch (response.status())
			{
				case CKR_OK:
					break;
				default:
					Genode::error("HSM replied with error ", Genode::Hex(response.status()));
					throw Hsm::Error_response_exception();
			}

			Payload const payload = response.payload_by_type(Data_type::typeDerSignature);
			if (!payload.is_valid()) {
				Genode::error("Signature data missing");
				throw Missing_payload_exception();
			}
			if (der_signature_size < payload.size()) {
				Genode::error("Signature buffer to small");
				throw Buffer_too_short_exception();
			}
			Genode::memcpy(der_signature_data, payload.data(), payload.size());
			return payload.size();

		}

		uint32_t sign(const Key_name& key_name,
		              const uint8_t* hash_data,
		              const uint32_t hash_size,
		              uint8_t* signature_data,
		              const uint32_t signature_size) const
		{
			if (!_login_success) {
				throw Not_logged_in { };
			}

			Message request(_heap, Command::CommandAsymmetricSingleSign);
			request.add(Data_type::typeCryptoMethod, Crypto_method::metEcDsaNone);
			request.add(Data_type::typeIdLabel, key_name.string());
			request.add(Data_type::typeSignData, hash_data, hash_size);

			Message const response = execute(request);

			switch (response.status())
			{
				case CKR_OK:
					break;
				default:
					Genode::error("HSM replied with error ", Genode::Hex(response.status()));
					throw Hsm::Error_response_exception();
			}

			Payload const signature = response.payload_by_type(Data_type::typeSignatureData);
			if (!signature.is_valid()) {
				Genode::error("Signature data missing");
				throw Missing_payload_exception();
			}

			uint32_t output_size = convert_signature(key_name,
			                                         signature.data(),
			                                         signature.size(),
			                                         signature_data,
			                                         signature_size);

			return output_size;
		}

		bool verify_signature(const Crypto_method method,
		                      const uint8_t* message,     const size_t message_size,
		                      const uint8_t* certificate, const size_t certificate_size,
		                      const uint8_t* signature,   const size_t signature_size) const
		{
			if (!_login_success) {
				throw Not_logged_in { };
			}

			Message request(_heap, Command::CommandVerifySignature);
			request.add(Data_type::typeCryptoMethod, method);
			request.add(Data_type::typeCertificateData, certificate, certificate_size);
			request.add(Data_type::typeSignData, message, message_size);
			request.add(Data_type::typeDerSignature, signature, signature_size);

			Message const response = execute(request);

			if (response.status() != CKR_OK) {
				Genode::error("HSM replied with error ", Genode::Hex(response.status()));
				return false;
			}

			return true;
		}

		/**
		 * Get the timestamp in seconds since epoch
		 * \result                    the timestamp in seconds since epoch
		 *
		 * \throw Error_response_exception
		 * \throw Buffer_too_short_exception
		 */
		uint64_t timestamp() const
		{
			if (!_login_success) {
				throw Not_logged_in { };
			}

			Message const request(_heap, Command::CommandSignTimestamp);
			Message const response = execute(request);

			switch (response.status())
			{
				case CKR_OK:
					break;
				default:
					Genode::error("HSM replied with error ", Genode::Hex(response.status()));
					throw Hsm::Error_response_exception();
			}

			Payload const payload = response.payload_by_type(Data_type::typeSecondsSinceEpoch);
			if (!payload.is_valid()) {
				Genode::error("SecondsSinceEpoch timestamp missing");
				throw Missing_payload_exception();
			}
			return payload.value<uint64_t>();
		}

		/**
		 * Get the timestamp in seconds since epoch and in RFC 3161 form
		 *
		 * \param key_name            the key name
		 * \param hash_buf            the hash for RFC 3161 timestamp signature
		 * \param hash_buf_size       the size of the hash
		 * \param rfc3161_buf         the out buffer for RFC 3161 timestamp
		 * \param rfc3161_buf_size    the size of the buffer for RFC 3161 timestamp
		 * \param rfc3161_length      the length of the RFC 3161 timestamp, 0 meaning invalid
		 * \result                    the timestamp in seconds since epoch
		 *
		 * \throw Error_response_exception
		 * \throw Buffer_too_short_exception
		 */
		uint64_t timestamp(Key_name const &key_name,
		                   uint8_t const  *hash_buf,
		                   size_t const    hash_buf_size,
		                   uint8_t        *rfc3161_buf,
		                   size_t const    rfc3161_buf_size,
		                   size_t         &rfc3161_length) const
		{
			if (!_login_success) {
				throw Not_logged_in { };
			}

			Message request(_heap, Command::CommandSignTimestamp);
			request.add(Data_type::typeIdLabel, key_name.string());
			request.add(Data_type::typeAlgorithmType, algSha256);
			request.add(Data_type::typeHashData, hash_buf, hash_buf_size);

			Message const response = execute(request);

			switch (response.status())
			{
				case CKR_OK:
					break;
				default:
					Genode::error("HSM replied with error ", Genode::Hex(response.status()));
					throw Hsm::Error_response_exception();
			}

			Payload const payload = response.payload_by_type(Data_type::typeSecondsSinceEpoch);
			if (!payload.is_valid()) {
				Genode::error("SecondsSinceEpoch timestamp missing");
				throw Missing_payload_exception();
			}
			uint64_t seconds = payload.value<uint64_t>();
			Payload const payload_rfc3161 = response.payload_by_type(Data_type::typeTimestampResponse);
			if (!payload_rfc3161.is_valid()) {
				// this is possible when the HSM doesn't have the rfc 3161 timestamp option enabled
				Genode::warning("rfc 3161 timestamp data missing");
				Genode::memset(rfc3161_buf, 0, rfc3161_buf_size);
				rfc3161_length = 0;
			} else {
				if (rfc3161_buf_size <payload_rfc3161.size()) {
					Genode::error("Timestamp buffer too small. needed size=", payload_rfc3161.size());
					throw Buffer_too_short_exception();
				}
				rfc3161_length = payload_rfc3161.size();
				Genode::memcpy(rfc3161_buf, payload_rfc3161.data(), rfc3161_length);
			}
			return seconds;
		}

		bool verify_timestamp_rfc_3161(const Key_name& key_name,
		                               const uint8_t* hash_data,
		                               const size_t hash_size,
		                               const uint8_t* timestamp_data,
		                               const size_t timestamp_size) const
		{
			if (!_login_success) {
				throw Not_logged_in { };
			}

			Message request(_heap, Command::CommandVerifyTimestamp);
			request.add(Data_type::typeIdLabel, key_name.string());
			request.add(Data_type::typeAlgorithmType, algSha256);
			request.add(Data_type::typeHashData, hash_data, hash_size);
			request.add(Data_type::typeTimestampResponse, timestamp_data, timestamp_size);

			Message const response = execute(request);

			if (response.status() != CKR_OK) {
				Genode::error("HSM replied with error ", Genode::Hex(response.status()));
				return false;
			}

			return true;
		}

		uint32_t cert_chain(const Key_name& key_name, uint8_t* cert_chain, const size_t size) const
		{
			if (!_login_success) {
				throw Not_logged_in { };
			}

			Message request(_heap, Command::CommandGetKeyValue);
			request.add(Data_type::typeIdLabel, key_name.string());
			request.add(Data_type::typeKeyFieldType, e_fieldId::fieldCertificateChain);
			request.add(Data_type::typeObjectType, e_objType::typeRksPrivateKey);

			Message const response = execute(request);

			if (response.status() != CKR_OK) {
				Genode::error("HSM replied with error ", Genode::Hex(response.status()));
				return false;
			}

			Payload const cert_chain_payload = response.payload_by_type(Data_type::typeKeyFieldValue);
			if (!cert_chain_payload.is_valid()) {
				Genode::error("Certificate chain data missing");
				throw Missing_payload_exception();
			}
			if (size < cert_chain_payload.size()) {
				Genode::error("Certificate chain buffer to small: expcected ", cert_chain_payload.size(), " actual ", size);
				throw Buffer_too_short_exception();
			}

			uint32_t const cert_chain_size = Genode::min(cert_chain_payload.size(), size);
			Genode::memcpy(cert_chain, cert_chain_payload.data(), cert_chain_size);
			return cert_chain_size;
		}

		/**
		 * Get the secure counter ID
		 *
		 * \result                    the ID of the counter. invalid if length==0
		 *
		 * \throw Error_response_exception
		 * \throw Buffer_too_short_exception
		 */
		template <size_t N>
		Genode::String<N> user_counter_id() const
		{
			if (!_login_success) {
				throw Not_logged_in { };
			}

			Message request(_heap, Command::CommandGetUserCounterSerial);
			Message const response = execute(request);

			if (response.status() != CKR_OK) {
				Genode::error("HSM replied with error ", Genode::Hex(response.status()));
				throw Hsm::Error_response_exception();
			}

			Genode::String<N> result;
			Payload const id = response.payload_by_type(Data_type::typeCounterId);
			if (!id.is_valid()) {
				// this is possible when the HSM doesn't have the counter option enabled
				Genode::warning("counter ID data missing");
			} else {
				if (N < id.size()) {
					Genode::error("counter ID buffer to small");
					throw Buffer_too_short_exception();
				}
				result = id.data<char>();
			}
			return result;
		}

		/**
		 * Get the secure counter value
		 *
		 * \param counter_value       the current counter value
		 * \result                    true if the counter value is valid
		 *
		 * \throw Error_response_exception
		 */
		bool get_counter_value(Genode::uint64_t &counter_value) const
		{
			if (!_login_success) {
				throw Not_logged_in { };
			}

			return _user_counter_value(counter_value, Counter_action::GetValue);
		}

		/**
		 * Increments and gets the incremented secure counter value
		 *
		 * \param counter_value       the incremented counter value
		 * \result                    true if the counter value is valid
		 *
		 * \throw Error_response_exception
		 */
		bool increment_counter_value(Genode::uint64_t &counter_value) const
		{
			if (!_login_success) {
				throw Not_logged_in { };
			}

			return _user_counter_value(counter_value, Counter_action::IncrementValue);
		}
};
