#pragma once

// Genode includes
#include <base/heap.h>
#include <libc/component.h>
#include <util/fifo.h>

// gapfruit includes
#include <util/hexdump.h>

// local includes
#include "command.h"
#include "exception.h"
#include "socket_wrapper.h"

namespace internal {
	template <typename DataType>
	class Payload;
}

namespace Hsm {
	struct __attribute__ ((packed)) Payload_header_short
	{
		uint16_t type;
		uint16_t length;
	};

	struct __attribute__ ((packed)) Payload_header_long
	{
		uint16_t type;
		uint16_t length;
		uint32_t long_length;
	};

	using Payload = internal::Payload<uint8_t const *>;

	struct Mutable_payload;
	struct Payload_helper;

} // namespace Hsm

struct Hsm::Payload_helper {

	static uint32_t padding_size(size_t payload_size)
	{
		payload_size += payload_size >= 0xFFFF ? sizeof(Payload_header_long) : sizeof(Payload_header_short);
		const uint32_t remainder = payload_size % sizeof(uint32_t);
		return (remainder == 0) ? 0 : (sizeof(uint32_t) - remainder);
	}

	static uint32_t padded_size(size_t payload_size)
	{
		payload_size += payload_size >= 0xFFFF ? sizeof(Payload_header_long) : sizeof(Payload_header_short);
		return payload_size + padding_size(payload_size);
	}
};

template <typename POINTER_TYPE>
class internal::Payload {
protected:
	POINTER_TYPE _data = nullptr;

public:
	Payload() = default;
	explicit Payload(POINTER_TYPE data) : _data(data) {}

	bool is_valid() const { return _data != nullptr; }

	constexpr Hsm::Payload_header_long const &header() const
	{
		return *reinterpret_cast<const Hsm::Payload_header_long*>(_data);
	}

	// size of the payload without header
	size_t size() const
	{
		return header().length != 0xFFFF ? header().length : header().long_length;
	}

	size_t header_size() const
	{
		return header().length != 0xFFFF ? sizeof(Hsm::Payload_header_short) : sizeof(Hsm::Payload_header_long);
	}

	Payload next() const
	{
		return Payload { _data + Hsm::Payload_helper::padded_size(size()) };
	}

	template <typename T = uint8_t>
	T const *data() const
	{
		return reinterpret_cast<T const *>(_data + header_size());
	}

	template <typename T>
	T const &value() const
	{
		return *reinterpret_cast<T const *>(_data + header_size());
	}

	uint8_t const *raw() const { return _data; }
};


struct Hsm::Mutable_payload : public internal::Payload<uint8_t *> {
	using internal::Payload<uint8_t *>::Payload;

	constexpr Payload_header_long &header()
	{
		return *reinterpret_cast<Payload_header_long*>(_data);
	}

	template <typename T = uint8_t>
	T *data()
	{
		return reinterpret_cast<T*>(_data + header_size());
	}

	size_t update(Data_type type, uint8_t const *payload, size_t size)
	{
		header().type = type;
		if (size < 0xFFFF) {
			header().length = static_cast<uint16_t>(size);
		} else {
			header().length = 0xFFFF;
			header().long_length = size;
		}
		memcpy(data(), payload, size);
		return Payload_helper::padded_size(size);
	}
};
